
###
### Copyright (c) 2004, 2005, David M. Blei and Jon D. McAuliffe.
### This software is distributed under the terms of the GNU
### General Public License.
###

make.conjugate.predictive.distribution <- function(state, data, data.sd)
{
  mean.mean <- mean(data);
  mean.sd   <- 4 * sd(data);

  counts <- table(state$classes);
  unique <- as.numeric(names(counts));
  counts <- as.vector(counts);
  
  alpha <- state$alpha;
  
  pd <- function (x)
  {
    likes <- numeric(length(counts) + 1);
    
    likes[1:length(counts)] <-
        sapply(unique,
               function (i)
               pred.density(x, data[state$classes==i],
                            mean.mean, mean.sd, data.sd));
      
    likes[length(counts) + 1] <-
        marg.density(x, mean.mean, mean.sd, data.sd)
    
    prior <- c(counts, alpha);
    prior <- prior/sum(prior);
    
    return(prior %*% likes);
  }

  return(function (x) { sapply(x, pd); });
}

bayesian.inference <- function(data, data.sd)
{
  state <- list();
  state$classes <- 1:length(data);
  state$alpha   <- (length(data) + 1) / (2 * log(length(data)));

  mean.mean <- mean(data);
  mean.sd   <- 4 * sd(data);
  
  for (i in 1:100)
  {
    cat("burn in iteration", i, "\n");	
    state <- dp.conj.gibbs(data = data, classes = state$classes,
                           alpha = state$alpha, mean.mean = mean.mean,
                           mean.sd = mean.sd, alpha.shape = 2,
                           alpha.scale = 2, data.sd = data.sd);
  }
  state.lst <- list();
  for (j in 1:200)
  {
    cat("iteration", j, "\n");	
    state <- dp.conj.gibbs(data = data, classes = state$classes,
                           alpha = state$alpha, mean.mean = mean.mean,
                           mean.sd = mean.sd, alpha.shape = 2,
                           alpha.scale = 2, data.sd = data.sd);
    state.lst[[j]] <- state;
  }
  return(state.lst);
}

# posterior sampling of \alpha from Escobar & West (1995)
# note that the scale parameter is called "rate" in rgamma

sample.alpha <- function(alpha, N, k, shape, scale)
{
  eta <- rbeta(1, alpha + 1, N);
  
  pi <- N * (scale - log(eta)) / (shape + k - 1);
  pi <- 1/(pi + 1);
  s <- rbinom(1, 1, pi);
  
  rgamma(1, shape = shape + k - (1 - s), rate = scale - log(eta));
}


# gibbs sampling under a conjugate prior
# algorithm 3 in Neal, 2000

dp.conj.gibbs <- function(data, classes, alpha, mean.mean,
                          mean.sd, alpha.shape, alpha.scale, data.sd)
{
  for (i in seq(along=data)) {

    if ((i %% 10) == 0) cat(i, "...", "\n")
    
    # get classes and counts
    counts <- table(classes);
    unique <- as.numeric(names(counts));
    counts <- as.vector(counts);
    
    k <- sum(counts>0);
    counts[classes[i]] <- counts[classes[i]] - 1;
    
    # account for possible removed class
    if (counts[classes[i]] == 0) {
      counts[classes[i]] <- counts[k];
      classes[classes == k] <- classes[i];
      counts[k] <- 0;
      k <- k - 1;
    }
    
    # compute likelihood of data under each class
    likelihood <- numeric(length(counts));

    likelihood <- numeric(k+1);
    for (j in 1:k) {
      indices <- (classes[-i] == j);
      likelihood[j] <- pred.density(data[i], data[-i][indices],
                                    mean.mean, mean.sd, data.sd);
    }

    likelihood[k+1] <- marg.density(data[i],
                                    prior.mean=mean.mean, prior.sd=mean.sd,
                                    data.sd=data.sd);
  
    # compute prior
    prior <- c(counts[1:k], alpha);
    
    # sample class
    classes[i] <- sample(1:length(prior), 1, prob = prior*likelihood);
  }
  
  # sample alpha
  counts <- table(classes);
  alpha  <- sample.alpha(alpha, length(data), length(counts),
                         alpha.shape, alpha.scale);
  
  # return state
  return(list(alpha = alpha, classes = classes));
}


# log predictive and marginal density in normal/normal model.
# notation from Bernardo and Smith (1992, p. 439)

pred.density <- function(x, obs, prior.mean, prior.sd, data.sd)
{
  lambda.0    <- 1/(prior.sd^2);
  lambda      <- 1/(data.sd^2);
  obs.mean    <- mean(obs);
  n           <- length(obs);
  lambda.n    <- lambda.0 + n * lambda; 
  mu.n        <- (lambda.0 * prior.mean + n * lambda * obs.mean) / lambda.n;
  lambda.pred <- lambda * lambda.n / (lambda.n + lambda);

  return(dnorm(x, mean = mu.n, sd = sqrt(1/lambda.pred)));
}

marg.density <- function(x, prior.mean, prior.sd, data.sd)
{
  lambda.0    <- 1/(prior.sd^2);
  lambda      <- 1/(data.sd^2);
  lambda.marg <- lambda * lambda.0 / (lambda.0 + lambda);
  
  return(dnorm(x, mean = prior.mean, sd = sqrt(1/lambda.marg)));
}
